###############################################################################
#
#   Agora Portfolio & Risk Management System
#
#   Copyright 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import urllib.request
import concurrent.futures

BASEURL = 'http://quotes.wsj.com/{0:s}/{1:s}'
TAGTSAMP = '<span class="timestamp_value" id="quote_dateTime">'
TAGPRICE = '<span id="quote_val">'
TAGDATA = '<span class="data_data">'
MAXTHREADS = 10


# -----------------------------------------------------------------------------
def fetchprc_wsj(ticker, country):
    """
    Description:
        Fetch real-time or delayed prices from Wall Street Journal.
    Inputs:
        country - country symbol as defined by Wall Street Journal
        ticker  - ticker symbol as defined by Wall Street Journal
    Returns:
        A list of (price, timestamp) tuples.
    """
    url = BASEURL.format(country, ticker)
    http = urllib.request.urlopen(url).read().decode("utf-8")

    # --- timestamp
    try:
        idx0 = http.index(TAGTSAMP) + len(TAGTSAMP)
        idx1 = http[idx0:].index('</span>')
        tstamp = http[idx0:idx0+idx1]
    except ValueError:
        if "Quote Not Found" in http:
            raise ValueError("incorrect country code or ticker: "
                             "{0:s}/{1:s}".format(country, ticker))
        else:
            print(url)
            raise

    # --- price
    try:
        idx0 = http.index(TAGPRICE) + len(TAGPRICE)
        idx1 = http[idx0:].index('</span>')
        prc = http[idx0:idx0+idx1].replace(",", "")
    except ValueError:
        if "Quote Not Found" in http:
            raise ValueError("incorrect country code or ticker: "
                             "{0:s}/{1:s}".format(country, ticker))
        else:
            raise

    return float(prc), str(tstamp)


# -----------------------------------------------------------------------------
def fetchdiv_wsj(ticker, country):
    """
    Description:
        Fetch current dividend info from Wall Street Journal.
    Inputs:
        country - country symbol as defined by Wall Street Journal
        ticker  - ticker symbol as defined by Wall Street Journal
    Returns:
        A tuple (ex dividend date, dividend).
    """
    url = BASEURL.format(country, ticker)
    http = urllib.request.urlopen(url).read().decode("utf-8")

    idx = http.find('<h3>Dividends</h3>') + len('<h3>Dividends</h3>')
    http = http[idx:]

    if '"msg_empty"' in http:
        return None

    idx = http.find('Latest payment') + len('Latest payment')
    http = http[idx:]
    idx = http.find('</span>') + len('</span>')
    http = http[idx:]
    idx0 = http.find(TAGDATA) + len(TAGDATA)
    idx1 = http.find('</span>')

    field = http[idx0:idx1]
    fidx0 = field.find('<small')
    fidx1 = field.find('</small>') + len('</small>')

    try:
        div = float(field[:fidx0] + field[fidx1:])
    except ValueError:
        raise ValueError("Wrong dividend value for "
                         "{0:s}: {1:s}".format(ticker, http[idx0:idx1]))

    idx = http.find('Ex-Dividend Date') + len('Ex-Dividend Date')
    http = http[idx:]
    idx0 = http.find(TAGDATA) + len(TAGDATA)
    idx1 = http[idx0:].find('</span>') + idx0
    exdt = http[idx0:idx1]

    return exdt, div


# -----------------------------------------------------------------------------
def fetchmany_wsj(symbols, target=fetchprc_wsj):
    with concurrent.futures.ThreadPoolExecutor(max_workers=MAXTHREADS) as exe:
        data = list(exe.map(lambda s: target(*s), symbols))
    return data


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    from onyx.core import TableFilter
    from datafeeds_config import BBG_MAPPING

    # --- remove rows that don't have a Google symbol set
    TableFilter(BBG_MAPPING, lambda row: row["WSJ"] is not None)

    # --- sort by bbg symbol
    BBG_MAPPING.sort(["Bloomberg"])

    data = fetchmany_wsj([tuple(row["WSJ"].split(":")) for row in BBG_MAPPING])
    for row in zip(BBG_MAPPING.column("Bloomberg"), data):
        print(row)



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date

from matplotlib.finance import fetch_historical_yahoo

import collections
import urllib.error
import socket
import time

__all__ = ["YahooError", "fetch_yahoo_hist"]


###############################################################################
class YahooError(Exception):
    pass


# -----------------------------------------------------------------------------
def fetch_yahoo_hist(ticker, start=None,
                     end=None, dividends=False, fail_after=3):
    """
    Description:
        Fetch historical stock prices or dividends from Yahoo Finance.
    Inputs:
        ticker     - ticker symbol as defined by Yahoo Finance
        start      - the start date
        end        - the end date
        dividends  - if True, return historical dividends
        fail_after - fail after this number of unsuccesseful attempts
    Returns:
        A list of quotes where each quote is a namedtuple.
    """
    # --- at times, the query to yahoo server fails, hence the multi attempt
    #     procedure used here.
    start = start or Date(1980, 1, 1)
    end = end or Date.today()

    for n in range(fail_after):
        try:
            data = fetch_historical_yahoo(
                    ticker, start, end, dividends=dividends)

        except (IOError, urllib.error.HTTPError, socket.error) as ex:
            if isinstance(ex, urllib.error.HTTPError) and ex.code == 404:
                raise YahooError(
                        "No available data for ticker '{0:s}' and "
                        "date range {1!s} - {2!s}".format(ticker, start, end))

            # --- sometimes it helps waiting before fetching prices again
            time.sleep(1.0)

        # --- cast datatypes and format data
        quotes = [row.rstrip().split(",") for row in data]
        header = [q.replace(" ", "_") for q in quotes[0]]
        quote_cls = collections.namedtuple("quote", header)
        quotes = [quote_cls(Date.parse(quote[0]),
                            *[float(q) for q in quote[1:]])
                  for quote in sorted(quotes[1:])]

        return quotes

    # --- nothing worked, raise an exception
    raise YahooError(
            "No response from yahoo server for ticker '{0:s}' "
            "and date range {1!s} - {2!s}".format(ticker, start, end))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################
"""
Library of functions and classes for using the bloomberg COM API v3. Derived
from the code written by Brian P. Smith (brian.p.smith@gmail.com).
"""
from onyx.core import Date, RDate, Curve, GCurve, HlocvCurve
from ..exceptions import DatafeedError, DatafeedFatal
from ..exceptions import SecurityError, FieldError

from . import com_helpers as bbgcom

import win32com.client
import pythoncom
import collections
import datetime
import sys

__all__ = ["BloombergApiInit", "BbgBDP", "BbgBDH"]

sec_err_fmt = "({0.security:s}, {0.category:s}, {0.message:s})"
fld_err_fmt = "({0.security:s}, {0.field:s}, {0.category:s}, {0.message:s}"


# -------------------------------------------------------------------------
def BloombergApiInit():
    pythoncom.CoInitialize()


###############################################################################
class ResponseHandler(object):
    # -------------------------------------------------------------------------
    def do_init(self, handler):
        """
        will be called prior to waiting for the message
        """
        self.waiting = True
        self.exc_info = None
        self.handler = handler

    # -------------------------------------------------------------------------
    def OnProcessEvent(self, evt):
        try:
            evt = win32com.client.CastTo(evt, "Event")
            if not self.handler:
                bbgcom.debug_event(evt)

            if evt.EventType == win32com.client.constants.RESPONSE:
                self.handler.on_event(evt, is_final=True)
                self.waiting = False
            elif evt.EventType == win32com.client.constants.PARTIAL_RESPONSE:
                self.handler.on_event(evt, is_final=False)
            else:
                self.handler.on_admin_event(evt)
        except:
            self.waiting = False
            self.exc_info = sys.exc_info()

    # -------------------------------------------------------------------------
    @property
    def has_deferred_exception(self):
        return self.exc_info is not None

    # -------------------------------------------------------------------------
    def raise_deferred_exception(self):
        raise self.exc_info[0].with_traceback(self.exc_info[1],
                                              self.exc_info[2])

    # -------------------------------------------------------------------------
    def do_cleanup(self):
        self.waiting = False
        self.exc_info = None
        self.handler = None


###############################################################################
class Request(object):
    # -------------------------------------------------------------------------
    def __init__(self, ignore_sec_err=False, ignore_fld_err=False):
        self.fld_errs = []
        self.sec_errs = []
        self.ignore_sec_err = ignore_sec_err
        self.ignore_fld_err = ignore_fld_err

    # -------------------------------------------------------------------------
    @property
    def has_exception(self):
        if not self.ignore_sec_err and len(self.sec_errs) > 0:
            return True
        if not self.ignore_fld_err and len(self.fld_errs) > 0:
            return True

    # -------------------------------------------------------------------------
    def raise_exception(self):
        if not self.ignore_sec_err and len(self.sec_errs) > 0:
            msgs = [sec_err_fmt.format(s) for s in self.sec_errs]
            raise SecurityError("{0:s}".format(",".join(msgs)))

        if not self.ignore_fld_err and len(self.fld_errs) > 0:
            msgs = [fld_err_fmt.format(s) for s in self.fld_errs]
            raise FieldError("{0:s}".format(",".join(msgs)))

        raise DatafeedError("Programmer Error: No exception to raise")

    # -------------------------------------------------------------------------
    def get_bbg_request(self, svc, session):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    def get_bbg_service_name(self):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    def on_event(self, evt, is_final):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    def on_admin_event(self, evt):
        pass

    # -------------------------------------------------------------------------
    def execute(self):
        session_name = "blpapicom.ProviderSession.1"
        session = win32com.client.DispatchWithEvents(session_name,
                                                     ResponseHandler)
        session.Start()
        try:
            svcname = self.get_bbg_service_name()
            if not session.OpenService(svcname):
                raise DatafeedFatal("failed to "
                                    "open service {0:s}".format(svcname))

            svc = session.GetService(svcname)
            asbbg = self.get_bbg_request(svc, session)
            session.SendRequest(asbbg)
            session.do_init(self)
            while session.waiting:
                pythoncom.PumpWaitingMessages()
            if session.has_deferred_exception:
                session.raise_deferred_exception()
            if self.has_exception:
                self.raise_exception()

        finally:
            session.Stop()
            session.do_cleanup()

        return self


###############################################################################
class ReferenceDataRequest(Request):
    # -------------------------------------------------------------------------
    def __init__(self, symbols, fields, overrides=None,
                 ignore_sec_err=False, ignore_fld_err=False):

        super().__init__(ignore_sec_err, ignore_fld_err)

        self.syms = [symbols] if isinstance(symbols, str) else symbols
        self.flds = [fields] if isinstance(fields, str) else fields
        self.overrides = overrides or {}

        # --- look for any date-override and convert it to its correct string
        #     representation
        for k, v in self.overrides.items():
            if isinstance(v, datetime.datetime):
                self.overrides[k] = v.strftime("%Y%m%d")

        # --- response related
        self.response = {}

    # -------------------------------------------------------------------------
    def get_bbg_service_name(self):
        return "//blp/refdata"

    # -------------------------------------------------------------------------
    def get_bbg_request(self, svc, session):
        # --- create the bloomberg request object
        req = svc.CreateRequest("ReferenceDataRequest")
        for sec in self.syms:
            req.GetElement("securities").AppendValue(sec)
        for fld in self.flds:
            req.GetElement("fields").AppendValue(fld)
        for field, value in self.overrides.items():
            ovd = req.GetElement("overrides").AppendElment()
            ovd.SetElement("fieldId", field)
            ovd.SetElement("value", value)
        return req

    # -------------------------------------------------------------------------
    def on_security_node(self, node):
        sid = bbgcom.get_child_value(node, "security")
        farr = node.GetElement("fieldData")
        fdata = bbgcom.get_child_values(farr, self.flds)
        if len(fdata) != len(self.flds):
            raise ValueError("field length must match data length")

        self.response[sid] = dict(zip(self.flds, fdata))

        ferrors = bbgcom.get_field_errors(node)
        if ferrors:
            self.fld_errs.extend(ferrors)

    # -------------------------------------------------------------------------
    def on_event(self, evt, is_final):
        """
        this is invoked from in response to
        COM PumpWaitingMessages - different thread
        """
        for msg in bbgcom.message_iter(evt):
            elm = msg.GetElement("securityData")
            for node, error in bbgcom.security_iter(elm):
                if error:
                    self.sec_errs.append(error)
                else:
                    self.on_security_node(node)


###############################################################################
class HistoricalDataRequest(Request):
    # -------------------------------------------------------------------------
    def __init__(self, symbols, fields,
                 start=None, end=None, period="DAILY", overrides=None,
                 ignore_sec_err=False, ignore_fld_err=False, **kwds):
        """Historical data request for bloomberg.

        Parameters
        ----------
        symbols : string or list
        fields : string or list
        start : start date (if None then use 1 year ago)
        end : end date (if None then use today)
        period : ("DAILY", "WEEKLY", "MONTHLY",
                  "QUARTERLY", "SEMI-ANNUAL", "YEARLY")
        overrides : dict
        ignore_fld_errs : bool
        ignore_sec_errs : bool
        """
        super().__init__(ignore_sec_err, ignore_fld_err)

        self.syms = [symbols] if isinstance(symbols, str) else symbols
        self.flds = [fields] if isinstance(fields, str) else fields
        self.start = start or Date.today() + RDate("-1y")
        self.end = end or Date.today()
        self.period = period
        self.overrides = overrides or {}
        self.kwds = kwds

        # --- look for any date-override and convert it to its correct string
        #     representation
        for k, v in self.overrides.items():
            if isinstance(v, datetime.datetime):
                self.overrides[k] = v.strftime("%Y%m%d")

        # --- response related
        self.response = {}

    # -------------------------------------------------------------------------
    def get_bbg_service_name(self):
        return "//blp/refdata"

    # -------------------------------------------------------------------------
    def get_bbg_request(self, svc, session):
        # --- create the bloomberg request object
        req = svc.CreateRequest("HistoricalDataRequest")
        for sec in self.syms:
            req.GetElement("securities").AppendValue(sec)
        for fld in self.flds:
            req.GetElement("fields").AppendValue(fld)
        req.Set("startDate", self.start.strftime("%Y%m%d"))
        req.Set("endDate", self.end.strftime("%Y%m%d"))
        req.Set("periodicitySelection", self.period)
        for field, value in self.overrides.items():
            ovd = req.GetElement("overrides").AppendElment()
            ovd.SetElement("fieldId", field)
            ovd.SetElement("value", value)
        for k, v in self.kwds.items():
            req.Set(k, v)
        return req

    # -------------------------------------------------------------------------
    def on_security_data_node(self, node):
        """
        process a securityData node - FIXME: currently not handling
        relateDate node
        """
        sid = bbgcom.get_child_value(node, "security")
        farr = node.GetElement("fieldData")

        dates, data = [], []
        for i in range(farr.NumValues):
            pt = farr.GetValue(i)
            dates.append(bbgcom.get_child_value(pt, "date"))
            values = [bbgcom.get_child_value(pt, f) for f in self.flds]
            data.append(dict(zip(self.flds, values)))

        self.response[sid] = GCurve(dates, data)

    # -------------------------------------------------------------------------
    def on_event(self, evt, is_final):
        """
        this is invoked from in response to
        COM PumpWaitingMessages - different thread
        """
        for msg in bbgcom.message_iter(evt):
            # Single security element in historical request
            node = msg.GetElement("securityData")
            if node.HasElement("securityError"):
                secid = bbgcom.get_child_value(node, "security")
                elm = node.GetElement("securityError")
                self.sec_errs.append(bbgcom.as_security_error(elm, secid))
            else:
                self.on_security_data_node(node)


###############################################################################
class IntrdayBarRequest(Request):
    # -------------------------------------------------------------------------
    def __init__(self, symbol, interval,
                 start=None, end=None, event="TRADE"):
        """Intraday bar request for bloomberg

        Parameters
        ----------
        symbols : string
        interval : number of minutes
        start : start date
        end : end date (if None then use today)
        event : (TRADE, BID, ASK, BEST_BID, BEST_ASK)
        """
        super().__init__()

        self.symbol = symbol
        self.interval = interval
        self.start = start or Date.today() + RDate("-30b")
        self.end = end or Date.today()
        self.event = event

        # --- response related
        self.response = collections.defaultdict(list)

    # -------------------------------------------------------------------------
    def get_bbg_service_name(self):
        return "//blp/refdata"

    # -------------------------------------------------------------------------
    def get_bbg_request(self, svc, session):
        # --- create the bloomberg request object
        start, end = self.start, self.end
        start = session.CreateDatetime(start.year, start.month,
                                       start.day, start.hour, start.minute)
        end = session.CreateDatetime(end.year, end.month,
                                     end.day, end.hour, end.minute)
        request = svc.CreateRequest("IntradayBarRequest")
        request.Set("security", self.symbol)
        request.Set("interval", self.interval)
        request.Set("eventType", self.event)
        request.Set("startDateTime", start)
        request.Set("endDateTime", end)
        return request

    # -------------------------------------------------------------------------
    def on_event(self, evt, is_final):
        """
        this is invoked from in response to
        COM PumpWaitingMessages - different thread
        """
        response = self.response

        for msg in bbgcom.message_iter(evt):
            bars = msg.GetElement("barData").GetElement("barTickData")
            for i in range(bars.NumValues):
                bar = bars.GetValue(i)
                ts = bar.GetElement(0).Value
                response["time"].append(Date(ts.year, ts.month,
                                             ts.day, ts.hour, ts.minute))
                response["value"].append({
                    "open": bar.GetElement(1).Value,
                    "high": bar.GetElement(2).Value,
                    "low": bar.GetElement(3).Value,
                    "close": bar.GetElement(4).Value,
                    "volume": bar.GetElement(5).Value,
                    "events": bar.GetElement(6).Value,
                })

        if is_final:
            self.response = GCurve(response["time"], response["value"])


# -----------------------------------------------------------------------------
def BbgBDP(securities, fields, overrides=None):
    """
    Description:
        Interface to a ReferenceDataRequest.
    Inputs:
        securities - one or more valid Bloomberg tickers
        fields     - one or more valid Bloomberg fields
        overrides  - an optional dictionary of overrides (or a tuple
                     of key, value tuples)
    Returns:
        A dictionary.
    """
    if overrides is not None and isinstance(overrides, tuple):
        overrides = dict(overrides)
    req = ReferenceDataRequest(securities, fields, overrides)
    req.execute()
    return req.response


# -----------------------------------------------------------------------------
def BbgBDH(security, fields, start, end, adj=True, overrides=None, **kwds):
    """
    Description:
        Interface to HistoricalDataRequest.
    Inputs:
        secirity  - a valid Bloomberg ticker
        fields    - one or more valid Bloomberg fields. "HLOCV" is a special
                    field type.
        start     - start date
        end       - end date
        adj       - True for dividend/split adjusted data
        overrides - an optional dictionary of overrides (or a tuple
                    of key, value tuples)
    Returns:
        A Curve.
    """
    if fields == "HLOCV":
        fields = ("PX_HIGH", "PX_LOW", "PX_OPEN", "PX_LAST", "VOLUME")
        is_hlocv = True
    else:
        is_hlocv = False
        fields = [fields] if isinstance(fields, str) else fields

    if overrides is not None and isinstance(overrides, tuple):
        overrides = dict(overrides)

    req = HistoricalDataRequest(security, fields, start, end,
                                overrides=overrides,
                                adjustmentNormal=adj,
                                adjustmentAbnormal=adj,
                                adjustmentSplit=adj, **kwds)
    req.execute()
    crv = req.response[security]

    if is_hlocv:
        if len(crv):
            # --- remove knots without a close value (should be extremely rare)
            knots = [(d, [v[fld] for fld in fields])
                     for d, v in crv if v["PX_LAST"] is not None]
            return HlocvCurve([d for d, v in knots], [v for d, v in knots])

        else:
            return HlocvCurve()

    elif len(fields) == 1:
        if len(crv):
            fld = fields[0]
            try:
                float(crv.front.value[fld])
            except ValueError:
                return crv
            else:
                return Curve(crv.dates, [val[fld] for val in crv.values])
        else:
            return Curve()
    else:
        return crv



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import urllib.request
import json
import concurrent.futures

__all__ = ["fetch_google", "fetchmany_google"]

BASEURL = "http://finance.google.com/finance/info?client=ig&q={0:s}:{1:s}"
MAXTHREADS = 10


# -----------------------------------------------------------------------------
def to_float(prc):
    return None if prc == "" else float(prc.replace(",", ""))


# -----------------------------------------------------------------------------
def fetch_google(exchange, ticker, fields=["l"]):
    """
    Description:
        Fetch real-time or delayed data from Google Finance.
    Inputs:
        exchange - exchange symbol as defined by Google Finance
        ticker   - ticker symbol as defined by Google Finance
        fields   - the list of fields to be returned
    Returns:
        A list of values (same order as fields).
    """
    url = BASEURL.format(exchange, ticker)
    http = urllib.request.urlopen(url).read().decode("cp1252")
    data = json.loads(http[3:])[0]
    return [data[field] for field in fields]


# -----------------------------------------------------------------------------
def fetchmany_google(symbols, target=fetch_google):
    with concurrent.futures.ThreadPoolExecutor(max_workers=MAXTHREADS) as exe:
        data = list(exe.map(lambda s: target(*s), symbols))
    return data



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Curve, DateRange

import urllib.request
import json
import pickle
import getpass
import os

BASE = "https://openexchangerates.org/api"

USER = getpass.getuser()
APP_ID = os.getenv("OPENEXCHANGERATES_APP_ID",
                   "8a439112054749a080ae72ce3114790c")
CACHE_FOLDER = os.getenv("FX_CACHE", os.path.join("/home", USER, "tmp"))


# -----------------------------------------------------------------------------
def __fetch_hist(date=None, app_id=APP_ID):
    if date is None:
        url = "{0:s}/latest.json?app_id={1:s}".format(BASE, app_id)
    else:
        url = "{0:s}/historical/{1:s}.json?"\
              "app_id={2:s}".format(BASE, date.strftime("%Y-%m-%d"), app_id)

    data = json.loads(urllib.request.urlopen(url).read().decode("utf-8"))
    del data["disclaimer"]
    del data["license"]
    return data


# -----------------------------------------------------------------------------
def get_rate(date, ccy, cache_folder=None):
    cache_folder = cache_folder or CACHE_FOLDER
    file_name = os.path.join(
            cache_folder, date.strftime("fx_cache_%Y-%m-%d.dat"))
    try:
        with open(file_name, "rb") as cache_file:
            data = pickle.load(cache_file)

    except FileNotFoundError:
        data = __fetch_hist(date)
        with open(file_name, "wb") as cache_file:
            pickle.dump(data, cache_file, -1)

    return data["rates"][ccy]


# -----------------------------------------------------------------------------
def fetch_historical(ccy, start, end, cache=None):
    """
    Description:
        Return historical prices for the ccy/USD cross, i.e. the value of
        1 unit of currency in US$ (sourced from www.openexchangerates.com).
    Inputs:
        ccy   - the currency crossed agains USD (ccy/USD)
        start - the start date
        end   - the end date
        cache - the folder used to cache files
    Returns:
        A Curve.
    """
    dts, vls = [], []
    for date in DateRange(start, end, "+1d"):
        try:
            value = 1.0 / get_rate(date, ccy, cache)
        except KeyError:
            pass
        else:
            dts.append(date)
            vls.append(value)

    if not len(dts):
        raise RuntimeError("No rates available for {0:s} in the "
                           "range {1!s} - {2!s}.".format(ccy, start, end))

    return Curve(dts, vls)



###############################################################################
#
#   Copyright: (c) 2015-2018 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Curve, Date, RDate

import urllib.request

__all__ = ["fetch_oanda_hist"]

BASE_URL = ("http://www.oanda.com/convert/fxhistory?"
            "date_fmt=us&date={0:8s}&date1={1:8s}&"
            "exch={2:3s}&expr=USD&exch2={2:3s}&expr2=USD&"
            "format=CSV&redirected=1")

RD_START = RDate("+A")
RD_END = RDate("+E")
RD_INC = RDate("+1d")


# -----------------------------------------------------------------------------
def fetch_oanda_hist(ccy, start, end):
    """
    Description:
        Return historical prices for the ccy/USD cross (from www.oanda.com).
    Inputs:
        ccy   - the currency crossed agains USD (ccy/USD)
        start - the start date
        end   - the end date
    Returns:
        A Curve.
    """
    dts, vls = [], []
    sd = ed = start

    # --- retrives time series in batches of 1 calendar year
    while ed < end:
        ed = min(end, ed + RD_END)
        sd = max(start, ed + RD_START)

        url = BASE_URL.format(ed.strftime("%m/%d/%y"),
                              sd.strftime("%m/%d/%y"), ccy)
        html = urllib.request.urlopen(url).read().decode("utf-8")
        idx0 = html.find("Daily averages:")
        html = html[idx0:]
        idx0 = html.find("<PRE>") + 5
        idx1 = html.find("</PRE>") - 1
        html = html[idx0:idx1]

        if len(html):
            for row in html.split("\n"):
                d, v = row.split(",")
                dts.append(Date.parse(d, False))
                vls.append(float(v))

        ed = ed + RD_INC

    return Curve(dts, vls)



###############################################################################
#
#   Copyright: (c) 2017 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from ..routerlib import DataRouter
from ..routerlib import RegistrationHandler, BbgBDPHandler, BbgBDHHandler

from onyx.core.utils.logging import setup_logging_from_json_config

import logging
import argh


# -----------------------------------------------------------------------------
def run(port=None, logging_config_file=None):
    setup_logging_from_json_config(logging_config_file)
    logger = logging.getLogger(__name__)
    handlers = [
        (r"/register/", RegistrationHandler),
        (r"/bbg-bdp/", BbgBDPHandler),
        (r"/bbg-bdh/", BbgBDHHandler),
    ]
    router = DataRouter(handlers=handlers, logger=logger)
    router.start(port)


# -----------------------------------------------------------------------------
def main():
    argh.dispatch_command(run)


# -----------------------------------------------------------------------------
#  for interactive use
if __name__ == "__main__":
    run()



###############################################################################
#
#   Copyright: (c) 2017 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import CurveField, HlocvCurveField
from onyx.core import DateOffset
from onyx.core import load_system_configuration

from .exceptions import DatafeedError, SecurityError, FieldError
from .exceptions import NoDataserverAvailable
from .utils import decode_message

import tornado.httpclient
import tornado.httputil

import datetime
import json

__all__ = ["DataClient"]


# -----------------------------------------------------------------------------
def to_upper(v):
    return v.upper() if isinstance(v, str) else v


###############################################################################
class DataClient(object):
    """
    This is a blocking client to send requests to a datafeed router.
    """
    # -------------------------------------------------------------------------
    def __init__(self, address=None, port=None, timeout=60):
        config = load_system_configuration()

        address = address or config.get("datafeed", "router_address")
        port = port or config.getint("datafeed", "router_port")

        self.bdp_url = "http://{0!s}:{1!s}/bbg-bdp/".format(address, port)
        self.bdh_url = "http://{0!s}:{1!s}/bbg-bdh/".format(address, port)
        self.timeout = timeout

    # -------------------------------------------------------------------------
    def BDP(self, sec, field,
            overrides=None, RT=False, fail_after=3):
        # --- ensure security, field, and overrides are upper case as the
        #     request caching is case sensitive
        sec = sec.upper()
        field = field.upper()

        if overrides is None:
            overrides = "null"
        else:
            # --- convert all keys and string-like values to upper case
            overrides = {k.upper(): to_upper(v) for k, v in overrides.items()}
            # --- look for any date-override and convert it to its correct
            #     string representation
            for k, v in overrides.items():
                if isinstance(v, datetime.datetime):
                    overrides[k] = v.strftime("%Y%m%d")
            overrides = json.dumps(overrides, sort_keys=True)

        request = {
            "sec": sec,
            "field": field,
            "overrides": overrides,
            "rt": RT,
            "type": "BDP",
        }

        response = self.query(self.bdp_url, request)

        return response["payload"]

    # -------------------------------------------------------------------------
    def BDH(self, sec, field, sd, ed,
            adj=True, overrides=None, fail_after=3):
        # --- ensure security, field, and overrides are upper case as the
        #     request caching is case sensitive
        sec = sec.upper()
        field = field.upper()

        if overrides is None:
            overrides = "null"
        else:
            # --- convert all keys and string-like values to upper case
            overrides = {k.upper(): to_upper(v) for k, v in overrides.items()}
            # --- look for any date-override and convert it to its correct
            #     string representation
            for k, v in overrides.items():
                if isinstance(v, datetime.datetime):
                    overrides[k] = v.strftime("%Y%m%d")
            overrides = json.dumps(overrides, sort_keys=True)

        request = {
            "sec": sec,
            "field": field,
            "start": sd.strftime("%Y%m%d"),
            "end": ed.strftime("%Y%m%d"),
            "adjusted": adj,
            "overrides": overrides,
            "type": "BDH",
        }

        response = self.query(self.bdh_url, request)

        if response["type"] == "Curve":
            return CurveField.from_json(None, response["payload"])
        elif response["type"] == "HlocvCurve":
            return HlocvCurveField.from_json(None, response["payload"])

    # -------------------------------------------------------------------------
    def HDP(self, sec, field, date, overrides=None):
        sd = DateOffset(date, "-1y")
        crv = self.BDH(sec, field, sd, date, adj=False, overrides=overrides)
        if len(crv):
            return crv.back.value
        else:
            return None

    # -------------------------------------------------------------------------
    def query(self, base_url, request):
        # --- send http GET request to datafeed router which then dispatches
        #     the request to one of the available datafeed servers
        url = tornado.httputil.url_concat(base_url, request)
        client = tornado.httpclient.HTTPClient()
        try:
            response = client.fetch(url, request_timeout=self.timeout)
        except tornado.httpclient.HTTPError as err:
            if err.code == 503:
                raise NoDataserverAvailable(str(err))
            else:
                raise
        finally:
            client.close()

        response = decode_message(response.body)
        self.manage_exceptions(response)
 
        return response

    # -------------------------------------------------------------------------
    def manage_exceptions(self, response):
        if response["type"] == "TimeoutError":
            raise TimeoutError(response["payload"])
        if response["type"] == "DatafeedError":
            raise DatafeedError(response["payload"])
        if response["type"] == "SecurityError":
            raise SecurityError(response["payload"])
        if response["type"] == "FieldError":
            raise FieldError(response["payload"])
        if response["type"] == "NotImplementedError":
            raise NotImplementedError(response["payload"])
